/*
 *  linux/drivers/block/loop.c
 *
 *  Written by Theodore Ts'o, 3/29/93
 * 
 * Copyright 1993 by Theodore Ts'o.  Redistribution of this file is
 * permitted under the GNU Public License.
 *
 * DES encryption plus some minor changes by Werner Almesberger, 30-MAY-1993
 * more DES encryption plus IDEA encryption by Nicholas J. Leon, June 20, 1996
 *
 * Modularized and updated for 1.1.16 kernel - Mitch Dsouza 28th May 1994
 * Adapted for 1.3.59 kernel - Andries Brouwer, 1 Feb 1996
 *
 * Fixed do_loop_request() re-entrancy - Vincent.Renardias@waw.com Mar 20, 1997
 *
 * Handle sparse backing files correctly - Kenn Humborg, Jun 28, 1998
 *
 * Loadable modules and other fixes by AK, 1998
 *
 * Make real block number available to downstream transfer functions, enables
 * CBC (and relatives) mode encryption requiring unique IVs per data block. 
 * Reed H. Petty, rhp@draper.net
 * 
 * Changing of backing file/device during operation for read-only loopback.
 * Jakub Jelinek <jakub@redhat.com>, Mar 31, 2000
 *
 * Added asynchron loopback requests and loopd.
 * Ingo Molnar <mingo@redhat.com>, Jul 10, 2000
 *
 *
 * Still To Fix:
 * - Advisory locking is ignored here. 
 * - Should use an own CAP_* category instead of CAP_SYS_ADMIN 
 * - Should use the underlying filesystems/devices read function if possible
 *   to support read ahead (and for write)
 *
 * WARNING/FIXME:
 * - The block number as IV passing to low level transfer functions is broken:
 *   it passes the underlying device's block number instead of the
 *   offset. This makes it change for a given block when the file is 
 *   moved/restored/copied and also doesn't work over NFS. 
 */ 

#define __KERNEL_SYSCALLS__
#include <linux/module.h>

#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/stat.h>
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/unistd.h>

#include <linux/init.h>

#include <asm/uaccess.h>

#include <linux/loop.h>
#include <linux/smp_lock.h>


#define MAJOR_NR LOOP_MAJOR

#define DEVICE_NAME "loop"
#define DEVICE_REQUEST do_lo_request
#define DEVICE_NR(device) (MINOR(device))
#define DEVICE_ON(device)
#define DEVICE_OFF(device)
#define DEVICE_NO_RANDOM
#define TIMEOUT_VALUE (6 * HZ)
#include <linux/blk.h>

#define MAX_LOOP 16
static struct loop_device loop_dev[MAX_LOOP];
static int loop_sizes[MAX_LOOP];
static int loop_blksizes[MAX_LOOP];

/*
 * Protect against deadlocking in ll_rw_block():
 */
#define MAX_LOOP_SEGMENTS (MAX_SEGMENTS/2)

static int max_loop_segments [MAX_LOOP] =
		 { [0 ... MAX_LOOP-1] = MAX_LOOP_SEGMENTS };
/*
 * Transfer functions
 */
static int transfer_none (struct loop_device *lo, int cmd, char *raw_buf,
		  		char *loop_buf, int size, int real_block)
{
	if (cmd == READ)
		memcpy(loop_buf, raw_buf, size);
	else
		memcpy(raw_buf, loop_buf, size);
	return 0;
}

static int transfer_xor (struct loop_device *lo, int cmd, char *raw_buf,
				 char *loop_buf, int size, int real_block)
{
	char	*in, *out, *key;
	int	i, keysize;

	if (cmd == READ) {
		in = raw_buf;
		out = loop_buf;
	} else {
		in = loop_buf;
		out = raw_buf;
	}
	key = lo->lo_encrypt_key;
	keysize = lo->lo_encrypt_key_size;
	for (i = 0; i < size; i++)
		*out++ = *in++ ^ key[(i & 511) % keysize];
	return 0;
}

static int none_status (struct loop_device *lo, struct loop_info *info)
{
	return 0; 
} 

static int xor_status (struct loop_device *lo, struct loop_info *info)
{
	if (info->lo_encrypt_key_size <= 0)
		return -EINVAL;
	return 0;
}

struct loop_func_table none_funcs = { 
	number: LO_CRYPT_NONE,
	transfer: transfer_none,
	init: none_status
}; 	

struct loop_func_table xor_funcs = { 
	number: LO_CRYPT_XOR,
	transfer: transfer_xor,
	init: xor_status
}; 	

/* xfer_funcs[0] is special - its release function is never called */ 
struct loop_func_table *xfer_funcs[MAX_LO_CRYPT] = {
	&none_funcs,
	&xor_funcs  
};

#define MAX_DISK_SIZE 1024*1024*1024

static int compute_loop_size (struct loop_device *lo,
				 struct dentry * lo_dentry, kdev_t lodev)
{
	if (S_ISREG(lo_dentry->d_inode->i_mode))
		return (lo_dentry->d_inode->i_size - lo->lo_offset) / BLOCK_SIZE;
	if (blk_size[MAJOR(lodev)])
		return blk_size[MAJOR(lodev)][MINOR(lodev)] -
			lo->lo_offset / BLOCK_SIZE;
	return MAX_DISK_SIZE;
}

static void figure_loop_size(struct loop_device *lo)
{
	loop_sizes[lo->lo_number] = compute_loop_size(lo, lo->lo_dentry, lo->lo_device);
}


#define BUG() ({ \
	__label__ __z; \
	printk("BUG at %p %d:%s!\n", &&__z, __LINE__, __FILE__);\
	*(int *) 0 = 0; \
	__z: \
})

static struct wait_queue *loopd_wait = NULL;
static struct wait_queue *loopd_done_wait = NULL;
static struct request *async_queue = NULL;

static int create_missing_block (struct loop_device *lo, int block, int blksize)
{
	struct file     *file;
	loff_t          new_offset;
	char            zero_buf[1] = { 0 };
	ssize_t         retval;
	mm_segment_t	old_fs;
	struct inode	*inode;

	file = lo->lo_backing_file;
	if (!file) {
		printk(KERN_WARNING "loop: cannot create block - no backing file\n");
		return 0;
	}

	if (!file->f_op) {
		printk(KERN_WARNING "loop: cannot create block - no file ops\n");
		return 0;
	}

	new_offset = block * blksize;

	if (file->f_op->llseek)
		file->f_op->llseek(file, new_offset, 0);
	else {
		/* Do what the default llseek() code would have done */
		file->f_pos = new_offset;
		file->f_reada = 0;
		file->f_version = ++global_event;
	}

	if (!file->f_op->write) {
		printk(KERN_WARNING "loop: cannot create block - file not writeable\n");
		return 0;
	}

	old_fs = get_fs();
	set_fs(get_ds());

	inode = file->f_dentry->d_inode;
	down(&inode->i_sem); 
	retval = file->f_op->write(file, zero_buf, 1, &file->f_pos);
	up(&inode->i_sem);
	
	set_fs(old_fs);

	if (retval < 0) {
		printk(KERN_WARNING "loop: cannot create block - FS write failed: code %d\n", 
		       (int)retval);
		return 0;
	} else
		return 1;
}

static void __do_lo_request (int final)
{
	int	real_block, block, offset, len, blksize, size, nr_bhs = 0, i;
	char	*dest_addr;
	struct loop_device *lo;
	struct buffer_head *bh, *bh_array[MAX_LOOP_SEGMENTS];
	struct request *current_request;
	int	block_present, rw;
	kdev_t	lo_device;

repeat:
	if (!final) {
		INIT_REQUEST;
		current_request = CURRENT;
		CURRENT = current_request->next;
	} else {
		struct request *prev = NULL;

		if (!async_queue)
			return;
		current_request = async_queue;
		/*
		 * Get to the last entry - we want FIFO fairness.
		 */
		while (current_request->next) {
			prev = current_request;
			current_request = current_request->next;
		}
		if (prev)
			prev->next = NULL;
		else
			async_queue = NULL;
	}
	rw = current_request->cmd;

	if (MINOR(current_request->rq_dev) >= MAX_LOOP)
		goto error_out;
	lo = &loop_dev[MINOR(current_request->rq_dev)];
	if (!lo->lo_dentry || !lo->transfer)
		goto error_out;

	blksize = BLOCK_SIZE;
	if (blksize_size[MAJOR(lo->lo_device)]) {
	    blksize = blksize_size[MAJOR(lo->lo_device)][MINOR(lo->lo_device)];
	    if (!blksize)
	      blksize = BLOCK_SIZE;
	}

	dest_addr = current_request->buffer;
	
	if (blksize < 512) {
		block = current_request->sector * (512/blksize);
		offset = 0;
	} else {
		block = current_request->sector / (blksize >> 9);
		offset = (current_request->sector % (blksize >> 9)) << 9;
	}
	block += lo->lo_offset / blksize;
	offset += lo->lo_offset % blksize;
	if (offset > blksize) {
		block++;
		offset -= blksize;
	}
	len = current_request->current_nr_sectors << 9;

	if (rw == WRITE) {
		if (lo->lo_flags & LO_FLAGS_READ_ONLY)
			goto error_out;
	} else if (rw != READ) {
		printk(KERN_ERR "unknown loop device command (%d)?!?", rw);
		goto error_out;
	}
	lo_device = lo->lo_device;

	spin_unlock_irq(&io_request_lock);
	down(&lo->sem);

	while (len > 0) {

		size = blksize - offset;
		if (size > len)
			size = len;

		real_block = block;
		block_present = 1;

		if (lo->lo_flags & LO_FLAGS_DO_BMAP) {
			real_block = bmap(lo->lo_dentry->d_inode, block);
			if (!real_block) {

				/*
				 * The backing file is a sparse file and
				 * this block doesn't exist.  If reading,
				 * return zeros. If writing, force the
				 * underlying FS to create the block
				 */
				if (rw == READ) {
					if (final)
						memset(dest_addr, 0, size);
					block_present = 0;
				} else {
					if (final)
						BUG();
					if (!create_missing_block(lo, block, blksize))
						goto error_out_lock;

					real_block = bmap(lo->lo_dentry->d_inode, block);
				}

			}
		}

		if (block_present) {
			bh = getblk(lo_device, real_block, blksize);
			if (!bh) {
				printk(KERN_ERR "loop: device %s: getblk(-, %d, %d) returned NULL",
					kdevname(lo_device),
					block, blksize);
				goto error_out_lock;
			}
			if (!buffer_uptodate(bh) &&
			    ((rw == READ) || (offset || (len < blksize))) ) {
				if (!final && (rw == READ))
					bh_array[nr_bhs++] = bh;
				else {
					if (rw == WRITE)
						ll_rw_block(READ, 1, &bh);
					wait_on_buffer(bh);
					if (!buffer_uptodate(bh)) {
						brelse(bh);
						brelse(bh);
						goto error_out_lock;
					}
				}
			}
 
			if (final || (rw == WRITE)) {
				if ((lo->transfer)(lo, rw, bh->b_data + offset,
						dest_addr, size, real_block)) {
					printk(KERN_ERR "loop: transfer error block %d\n", block);
					if (final)
						brelse(bh);
					brelse(bh);
					goto error_out_lock;
				}

				if (rw == WRITE) {
					up(&lo->sem);
					mark_buffer_uptodate(bh, 1);
					mark_buffer_dirty(bh, 1);
					down(&lo->sem);
				}
				if (final)
					/*
					 * Twice because the startup path
					 * did a getblk as well.
					 */
					brelse(bh);
				brelse(bh);
			}
		}
		dest_addr += size;
		len -= size;
		offset = 0;
		block++;
	}
	if (nr_bhs) {
		if (rw == WRITE)
			BUG();
		ll_rw_block(READ, nr_bhs, bh_array);
	}

	spin_lock_irq(&io_request_lock);
	if (final || (rw == WRITE)) {
		current_request->next = CURRENT;
		CURRENT = current_request;
		end_request(1);
	} else {
		current_request->next = async_queue;
		async_queue = current_request;
	}
	up(&lo->sem);
	goto repeat;
error_out_lock:
	if (nr_bhs && final)
		BUG();
	for (i = 0; i < nr_bhs; i++)
		brelse(bh_array[i]);
	up(&lo->sem);
	spin_lock_irq(&io_request_lock);
	goto end_req;
error_out:
	if (final)
		BUG();
end_req:
	current_request->next = CURRENT;
	CURRENT = current_request;
	end_request(0);
	goto repeat;
}

static void do_lo_request (void)
{
	__do_lo_request(0);
	wake_up(&loopd_wait);
	wake_up(&loopd_done_wait);
}

static void __wait_loopd (struct loop_device *lo)
{
	struct wait_queue wait = { current, NULL };

	add_wait_queue(&loopd_done_wait, &wait);
repeat:
	current->state = TASK_UNINTERRUPTIBLE;

	if (async_queue) {
		spin_unlock_irq(&io_request_lock);
		up(&lo->sem);
		schedule();
		down(&lo->sem);
		spin_lock_irq(&io_request_lock);
		goto repeat;
	}
	current->state = TASK_RUNNING;

	remove_wait_queue(&loopd_done_wait, &wait);
}

static int loop_change_fd(struct loop_device *lo, struct file *lo_file, kdev_t dev, unsigned int arg, int set)
{
	struct file	*file;
	struct inode	*inode;
	kdev_t		lo_device;
	int		lo_flags;
	struct file	*lo_backing_file = NULL;
	int error;

	if (set) {
		MOD_INC_USE_COUNT;

		error = -EBUSY;
		if (lo->lo_dentry)
			goto out;
	} else {
		/* Changing backing file is only allowed for read-only devices now.  */
		error = -EINVAL;
		if (!lo->lo_dentry || !(lo->lo_flags & LO_FLAGS_READ_ONLY))
			goto out;
	}

	error = -EBADF;
	file = fget(arg);
	if (!file)
		goto out;

	error = -EINVAL;
	inode = file->f_dentry->d_inode;
	if (!inode) {
		printk(KERN_ERR "loop_change_fd: NULL inode?!?\n");
		goto out_putf;
	}

	lo_device = inode->i_rdev;
	lo_flags = 0;
	/* Backed by a block device - don't need to hold onto
	   a file structure */
	lo_backing_file = NULL;

	if (S_ISBLK(inode->i_mode)) {
		error = blkdev_open(inode, file);
	} else if (S_ISREG(inode->i_mode)) {
		if (!inode->i_op->bmap) { 
			printk(KERN_ERR "loop: device has no block access/not implemented\n");
			goto out_putf;
		}

		/* Backed by a regular file - we need to hold onto
		   a file structure for this file.  We'll use it to
		   write to blocks that are not already present in 
		   a sparse file.  We create a new file structure
		   based on the one passed to us via 'arg'.  This is
		   to avoid changing the file structure that the
		   caller is using */

		lo_device = inode->i_dev;
		lo_flags = LO_FLAGS_DO_BMAP;

		error = -ENFILE;
		lo_backing_file = get_empty_filp();
		if (lo_backing_file) {
			lo_backing_file->f_mode = file->f_mode;
			lo_backing_file->f_pos = file->f_pos;
			lo_backing_file->f_flags = file->f_flags;
			lo_backing_file->f_owner = file->f_owner;
			lo_backing_file->f_dentry = file->f_dentry;
			lo_backing_file->f_op = file->f_op;
			lo_backing_file->private_data = file->private_data;

			error = get_write_access(inode);
			if (error) {
				put_filp(lo_backing_file);
				lo_backing_file = NULL;
			}
		}
	} else
		goto out_putf;
	if (error)
		goto out_putf;

	if (set) {
		if (IS_RDONLY (inode) || is_read_only(lo_device)
		    || !(lo_file->f_mode & FMODE_WRITE))
			lo_flags |= LO_FLAGS_READ_ONLY;
		else
			invalidate_inode_pages (inode);

		set_device_ro(dev, (lo_flags & LO_FLAGS_READ_ONLY) != 0);

		lo->lo_device = lo_device;
		lo->lo_flags = lo_flags;
		lo->lo_backing_file = lo_backing_file;
		lo->lo_dentry = dget(file->f_dentry);
		lo->transfer = NULL;
		lo->ioctl = NULL;
		figure_loop_size(lo);
	} else {
		int size = compute_loop_size(lo, file->f_dentry, lo_device);
		kdev_t old_device;
		struct dentry *old_dentry;
		struct file *old_backing_file;

		error = -EINVAL;
		if (size != loop_sizes[lo->lo_number])
			/* Loop size has changed. Don't allow it. */
			goto out_put_all;

		/*
		 * Wait for loopd IO to drain and change the fd in
		 * an atomic way, under the io_request spinlock.
		 */

		down(&lo->sem);
		spin_lock_irq(&io_request_lock);

		if (async_queue)
			__wait_loopd(lo);
		if (async_queue)
			BUG();
		old_device = lo->lo_device;
		old_backing_file = lo->lo_backing_file;
		old_dentry = lo->lo_dentry;
		lo->lo_device = lo_device;
		lo->lo_backing_file = lo_backing_file;
		lo->lo_dentry = dget(file->f_dentry);
		lo->lo_flags = lo_flags | LO_FLAGS_READ_ONLY;

		spin_unlock_irq(&io_request_lock);
		up(&lo->sem);

		if (S_ISBLK(old_dentry->d_inode->i_mode))
			blkdev_release(old_dentry->d_inode);
		if (old_backing_file)
			fput(old_backing_file);
		else
			dput(old_dentry);
		error = 0;
		goto out_putf;
out_put_all:
		if (S_ISBLK(inode->i_mode))
			blkdev_release(inode);
	}
out_putf:
	if (error && lo_backing_file)
		put_filp(lo_backing_file);
	fput(file);
out:
	if (error && set)
		MOD_DEC_USE_COUNT;
	return error;
}

static int loop_release_xfer (struct loop_device *lo)
{
	int err = 0; 
	if (lo->lo_encrypt_type) {
		struct loop_func_table *xfer= xfer_funcs[lo->lo_encrypt_type]; 
		if (xfer && xfer->release)
			err = xfer->release(lo); 
		if (xfer && xfer->unlock)
			xfer->unlock(lo); 
		lo->lo_encrypt_type = 0;
	}
	return err;
}

static int loop_init_xfer (struct loop_device *lo, int type,struct loop_info *i)
{
	int err = 0; 
	if (type) {
		struct loop_func_table *xfer = xfer_funcs[type]; 
		if (xfer->init)
			err = xfer->init(lo, i);
		if (!err) { 
			lo->lo_encrypt_type = type;
			if (xfer->lock)
				xfer->lock(lo);
		}
	}
	return err;
}  

static int loop_clr_fd (struct loop_device *lo, kdev_t dev)
{
	struct dentry *dentry = lo->lo_dentry;

	if (!dentry)
		return -ENXIO;
	if (lo->lo_refcnt > 1)	/* we needed one fd for the ioctl */
		return -EBUSY;

	if (S_ISBLK(dentry->d_inode->i_mode))
		blkdev_release (dentry->d_inode);
	lo->lo_dentry = NULL;

	if (lo->lo_backing_file) {
		fput(lo->lo_backing_file);
		lo->lo_backing_file = NULL;
	} else
		dput(dentry);

	loop_release_xfer(lo);
	lo->transfer = NULL;
	lo->ioctl = NULL;
	lo->lo_device = 0;
	lo->lo_encrypt_type = 0;
	lo->lo_offset = 0;
	lo->lo_encrypt_key_size = 0;
	memset(lo->lo_encrypt_key, 0, LO_KEY_SIZE);
	memset(lo->lo_name, 0, LO_NAME_SIZE);
	loop_sizes[lo->lo_number] = 0;
	invalidate_buffers(dev);
	MOD_DEC_USE_COUNT;
	return 0;
}

static int loop_set_status (struct loop_device *lo, struct loop_info *arg)
{
	struct loop_info info; 
	int err;
	unsigned int type;

	if (lo->lo_encrypt_key_size && lo->lo_key_owner != current->uid && 
	    !capable(CAP_SYS_ADMIN))
		return -EPERM;
	if (!lo->lo_dentry)
		return -ENXIO;
	if (copy_from_user(&info, arg, sizeof (struct loop_info)))
		return -EFAULT; 
	if ((unsigned int) info.lo_encrypt_key_size > LO_KEY_SIZE)
		return -EINVAL;
	type = info.lo_encrypt_type; 
	if (type >= MAX_LO_CRYPT || !xfer_funcs[type])
		return -EINVAL;
	err = loop_release_xfer(lo);
	if (!err) 
		err = loop_init_xfer(lo, type, &info);
	if (err)
		return err;	

	lo->lo_offset = info.lo_offset;
	strncpy(lo->lo_name, info.lo_name, LO_NAME_SIZE);

	lo->transfer = xfer_funcs[type]->transfer;
	lo->ioctl = xfer_funcs[type]->ioctl;
	lo->lo_encrypt_key_size = info.lo_encrypt_key_size;
	lo->lo_init[0] = info.lo_init[0];
	lo->lo_init[1] = info.lo_init[1];
	if (info.lo_encrypt_key_size) {
		memcpy(lo->lo_encrypt_key, info.lo_encrypt_key, 
		       info.lo_encrypt_key_size);
		lo->lo_key_owner = current->uid; 
	}	
	figure_loop_size(lo);
	return 0;
}

static int loop_get_status (struct loop_device *lo, struct loop_info *arg)
{
	struct loop_info	info;

	if (!lo->lo_dentry)
		return -ENXIO;
	if (!arg)
		return -EINVAL;
	memset(&info, 0, sizeof(info));
	info.lo_number = lo->lo_number;
	info.lo_device = kdev_t_to_nr(lo->lo_dentry->d_inode->i_dev);
	info.lo_inode = lo->lo_dentry->d_inode->i_ino;
	info.lo_rdevice = kdev_t_to_nr(lo->lo_device);
	info.lo_offset = lo->lo_offset;
	info.lo_flags = lo->lo_flags;
	strncpy(info.lo_name, lo->lo_name, LO_NAME_SIZE);
	info.lo_encrypt_type = lo->lo_encrypt_type;
	if (lo->lo_encrypt_key_size && capable(CAP_SYS_ADMIN)) {
		info.lo_encrypt_key_size = lo->lo_encrypt_key_size;
		memcpy(info.lo_encrypt_key, lo->lo_encrypt_key,
		       lo->lo_encrypt_key_size);
	}
	return copy_to_user(arg, &info, sizeof(info)) ? -EFAULT : 0;
}

static int lo_ioctl (struct inode * inode, struct file * file,
			unsigned int cmd, unsigned long arg)
{
	struct loop_device *lo;
	int dev;

	if (!inode)
		return -EINVAL;
	if (MAJOR(inode->i_rdev) != MAJOR_NR) {
		printk(KERN_WARNING "lo_ioctl: pseudo-major != %d\n", MAJOR_NR);
		return -ENODEV;
	}
	dev = MINOR(inode->i_rdev);
	if (dev >= MAX_LOOP)
		return -ENODEV;
	lo = &loop_dev[dev];
	switch (cmd) {
	case LOOP_SET_FD:
		return loop_change_fd(lo, file, inode->i_rdev, arg, 1);
	case LOOP_CLR_FD:
		return loop_clr_fd(lo, inode->i_rdev);
	case LOOP_CHANGE_FD:
		return loop_change_fd(lo, file, inode->i_rdev, arg, 0);
	case LOOP_SET_STATUS:
		return loop_set_status(lo, (struct loop_info *) arg);
	case LOOP_GET_STATUS:
		return loop_get_status(lo, (struct loop_info *) arg);
	case BLKGETSIZE:   /* Return device size */
		if (!lo->lo_dentry)
			return -ENXIO;
		if (!arg)
			return -EINVAL;
		return put_user(loop_sizes[lo->lo_number] << 1, (long *) arg);
	default:
		return lo->ioctl ? lo->ioctl(lo, cmd, arg) : -EINVAL;
	}
	return 0;
}

static int lo_open (struct inode *inode, struct file *file)
{
	struct loop_device *lo;
	int	dev, type;

	if (!inode)
		return -EINVAL;
	if (MAJOR(inode->i_rdev) != MAJOR_NR) {
		printk(KERN_WARNING "lo_open: pseudo-major != %d\n", MAJOR_NR);
		return -ENODEV;
	}
	dev = MINOR(inode->i_rdev);
	if (dev >= MAX_LOOP)
		return -ENODEV;
	lo = &loop_dev[dev];

	type = lo->lo_encrypt_type; 
	if (type && xfer_funcs[type] && xfer_funcs[type]->lock)
		xfer_funcs[type]->lock(lo);
	lo->lo_refcnt++;
	MOD_INC_USE_COUNT;
	return 0;
}

static int lo_release (struct inode *inode, struct file *file)
{
	struct loop_device *lo;
	int	dev, err;

	if (!inode)
		return 0;
	if (MAJOR(inode->i_rdev) != MAJOR_NR) {
		printk(KERN_WARNING "lo_release: pseudo-major != %d\n", MAJOR_NR);
		return 0;
	}
	dev = MINOR(inode->i_rdev);
	if (dev >= MAX_LOOP)
		return 0;
	err = fsync_dev(inode->i_rdev);
	lo = &loop_dev[dev];
	if (lo->lo_refcnt <= 0)
		printk(KERN_ERR "lo_release: refcount(%d) <= 0\n", lo->lo_refcnt);
	else  {
		int type  = lo->lo_encrypt_type;
		--lo->lo_refcnt;
		if (xfer_funcs[type] && xfer_funcs[type]->unlock)
			xfer_funcs[type]->unlock(lo);
		MOD_DEC_USE_COUNT;
	}
	return err;
}

static struct file_operations lo_fops = {
	NULL,			/* lseek - default */
	block_read,		/* read - general block-dev read */
	block_write,		/* write - general block-dev write */
	NULL,			/* readdir - bad */
	NULL,			/* poll */
	lo_ioctl,		/* ioctl */
	NULL,			/* mmap */
	lo_open,		/* open */
	NULL,			/* flush */
	lo_release		/* release */
};

/*
 * And now the modules code and kernel interface.
 */
#ifdef MODULE
#define loop_init init_module
#endif

int loop_register_transfer (struct loop_func_table *funcs)
{
	if ((unsigned)funcs->number > MAX_LO_CRYPT || xfer_funcs[funcs->number])
		return -EINVAL;
	xfer_funcs[funcs->number] = funcs;
	return 0; 
}

int loop_unregister_transfer (int number)
{
	struct loop_device *lo; 

	if ((unsigned)number >= MAX_LO_CRYPT)
		return -EINVAL; 
	for (lo = &loop_dev[0]; lo < &loop_dev[MAX_LOOP]; lo++) { 
		int type = lo->lo_encrypt_type;
		if (type == number) { 
			xfer_funcs[type]->release(lo);
			lo->transfer = NULL; 
			lo->lo_encrypt_type = 0; 
		}
	}
	xfer_funcs[number] = NULL; 
	return 0; 
}

EXPORT_SYMBOL(loop_register_transfer);
EXPORT_SYMBOL(loop_unregister_transfer);

static int loopd (void *unused)
{
	struct wait_queue wait = { current, NULL };

	lock_kernel();
	/* printk("loopback system thread started ...\n"); */
	strcpy(current->comm, "loopd");
	init_waitqueue(&loopd_wait);
	init_waitqueue(&loopd_done_wait);

	spin_lock_irq(&io_request_lock);
repeat:
	__do_lo_request(1);

	current->state = TASK_UNINTERRUPTIBLE;
	add_wait_queue(&loopd_wait, &wait);

	if (!async_queue) {
		spin_unlock_irq(&io_request_lock);
		wake_up(&loopd_done_wait);
		schedule();
		spin_lock_irq(&io_request_lock);
	}

	remove_wait_queue(&loopd_wait, &wait);
	current->state = TASK_RUNNING;

	goto repeat;

	return 0;
}

int __init loop_init (void) 
{
	int	i;

	kernel_thread(loopd, NULL, CLONE_FS | CLONE_FILES | CLONE_SIGHAND);

	max_segments[MAJOR_NR] = max_loop_segments;

	if (register_blkdev(MAJOR_NR, "loop", &lo_fops)) {
		printk(KERN_WARNING "Unable to get major number %d for loop device\n",
		       MAJOR_NR);
		return -EIO;
	}
#ifndef MODULE
	printk(KERN_INFO "loop: registered device at major %d\n", MAJOR_NR);
#endif

	blk_dev[MAJOR_NR].request_fn = DEVICE_REQUEST;
	for (i = 0; i < MAX_LOOP; i++) {
		memset(&loop_dev[i], 0, sizeof(struct loop_device));
		loop_dev[i].lo_number = i;
		loop_dev[i].sem = MUTEX;
	}
	memset(&loop_sizes, 0, sizeof(loop_sizes));
	memset(&loop_blksizes, 0, sizeof(loop_blksizes));
	blk_size[MAJOR_NR] = loop_sizes;
	blksize_size[MAJOR_NR] = loop_blksizes;

	return 0;
}

#ifdef MODULE
void cleanup_module (void) 
{
	if (unregister_blkdev(MAJOR_NR, "loop"))
		printk(KERN_WARNING "loop: cannot unregister blkdev\n");
}
#endif
