/*
 * iobuf.c
 *
 * Keep track of the general-purpose IO-buffer structures used to track
 * abstract kernel-space io buffers.
 * 
 */

#include <linux/iobuf.h>
#include <linux/malloc.h>
#include <linux/slab.h>
#include <linux/bigmem.h>

static kmem_cache_t *kiobuf_cachep;

void __init kiobuf_init(void)
{
	kiobuf_cachep =  kmem_cache_create("kiobuf",
					   sizeof(struct kiobuf),
					   0,
					   SLAB_HWCACHE_ALIGN, NULL, NULL);
	if(!kiobuf_cachep)
		panic("Cannot create kernel iobuf cache\n");
}


int alloc_kiovec(int nr, struct kiobuf **bufp)
{
	int i;
	struct kiobuf *iobuf;
	
	for (i = 0; i < nr; i++) {
		iobuf = kmem_cache_alloc(kiobuf_cachep, SLAB_KERNEL);
		if (!iobuf) {
			free_kiovec(i, bufp);
			return -ENOMEM;
		}
		
		memset(iobuf, 0, sizeof(*iobuf));
		iobuf->array_len  = KIO_STATIC_PAGES;
		iobuf->pagelist   = iobuf->page_array;
		iobuf->maplist    = iobuf->map_array;
		iobuf->bouncelist = iobuf->bounce_array;
		*bufp++ = iobuf;
	}
	
	return 0;
}

void clear_kiobuf_bounce_pages(struct kiobuf *iobuf)
{
	int i;
	
	if (!iobuf->bounced)
		return;
	
	for (i = 0; i < iobuf->nr_pages; i++) {
		unsigned long page = iobuf->bouncelist[i];
		if (page)
			free_page(page);
	}
	iobuf->bounced = 0;
}

void free_kiovec(int nr, struct kiobuf **bufp) 
{
	struct kiobuf *iobuf;
	int i;
	
	for (i = 0; i < nr; i++) {
		iobuf = bufp[i];
		clear_kiobuf_bounce_pages(iobuf);
		if (iobuf->array_len > KIO_STATIC_PAGES) {
			kfree (iobuf->pagelist);
		}
		kmem_cache_free(kiobuf_cachep, bufp[i]);
	}
}

int expand_kiobuf(struct kiobuf *iobuf, int wanted)
{
	unsigned long *	pagelist, * bouncelist;
	struct page ** maplist;
	
	if (iobuf->array_len >= wanted)
		return 0;

	/*
	 * kmalloc enough space for the page, map and bounce lists all
	 * at once. 
	 */
	pagelist = (unsigned long *) 
		kmalloc(3 * wanted * sizeof(unsigned long), GFP_KERNEL);
	if (!pagelist)
		return -ENOMEM;
	
	/* Did it grow while we waited? */
	if (iobuf->array_len >= wanted) {
		kfree(pagelist);
		return 0;
	}

	maplist    = (struct page **) (pagelist + wanted);
	bouncelist = pagelist + 2 * wanted;

	memcpy (pagelist, iobuf->pagelist,
		iobuf->array_len * sizeof(unsigned long));
	memcpy (maplist, iobuf->maplist,
		iobuf->array_len * sizeof(struct page **));
	memcpy (bouncelist, iobuf->bouncelist,
		iobuf->array_len * sizeof(unsigned long));

	if (iobuf->array_len > KIO_STATIC_PAGES) {
		kfree (iobuf->pagelist);
	}
	
	iobuf->pagelist   = pagelist;
	iobuf->maplist    = maplist;
	iobuf->bouncelist = bouncelist;
	iobuf->array_len  = wanted;
	return 0;
}

/*
 * Test whether a given page from the bounce buffer matches the given
 * gfp_mask.  Return true if a bounce buffer is required for this
 * page. 
 */

static inline int test_bounce_page(unsigned long page, 
				   struct page * map,
				   int gfp_mask)
{
	/* Unmapped pages from PCI memory or BIGMEM pages always need a
	 * bounce buffer unless the caller is prepared to accept
	 * GFP_BIGMEM pages. */
	
	if (!map || PageBIGMEM(map) )
		/* Careful, the following must return the right value
		 * even if CONFIG_BIGMEM is not set */
		return !(gfp_mask & __GFP_BIGMEM);
	
	/* A DMA-able page never needs a bounce buffer */
	if (PageDMA(map))
		return 0;
	
	/* Otherwise it is a non-ISA-DMA-capable page and needs bounce
	 * buffers if GFP_DMA is requested */
	return gfp_mask & __GFP_DMA;
}

int setup_kiobuf_bounce_pages(struct kiobuf *iobuf, int gfp_mask)
{
	int i;
	
	clear_kiobuf_bounce_pages(iobuf);
	
	for (i = 0; i < iobuf->nr_pages; i++) {
		struct page *map = iobuf->maplist[i];
		unsigned long page = iobuf->pagelist[i];
		unsigned long bounce_page;
		
		if (!test_bounce_page(page, map, gfp_mask)) {
			iobuf->bouncelist[i] = 0;
			continue;
		}
		
		bounce_page = __get_free_page(gfp_mask);
		if (!bounce_page)
			goto error;

		iobuf->bouncelist[i] = bounce_page;
		iobuf->bounced = 1;
	}
	return 0;
	
 error:
	clear_kiobuf_bounce_pages(iobuf);
	return -ENOMEM;
}

/*
 * Copy a bounce buffer.  For completion of partially-failed read IOs,
 * we need to be able to place an upper limit on the data successfully
 * transferred from bounce buffers to the user's own buffers.  
 */

void kiobuf_copy_bounce(struct kiobuf *iobuf, int direction, int max)
{
	int i;
	int offset, length;
	
	if (!iobuf->bounced)
		return;
	
	offset = iobuf->offset;
	length = iobuf->length;
	if (max >= 0 && length > max)
		length = max;
	
	i = 0;

	if (offset > PAGE_SIZE) {
		i = (offset >> PAGE_SHIFT);
		offset &= ~PAGE_MASK;
	}
	
	for (; i < iobuf->nr_pages && length > 0; i++) {
		unsigned long page = iobuf->pagelist[i];
		unsigned long bounce_page = iobuf->bouncelist[i];
		unsigned long kin, kout;
		int pagelen = length;
		
		if (bounce_page) {
			if ((pagelen+offset) > PAGE_SIZE)
				pagelen = PAGE_SIZE - offset;
		
			if (direction == COPY_TO_BOUNCE) {
				kin  = kmap(page, KM_READ);
				kout = kmap(bounce_page, KM_WRITE);
			} else {
				kin  = kmap(bounce_page, KM_READ);
				kout = kmap(page, KM_WRITE);
			}
			
			memcpy((char *) (kout+offset), 
			       (char *) (kin+offset),
			       pagelen);
			kunmap(kout, KM_WRITE);
			kunmap(kin, KM_READ);
		}
		
		length -= pagelen;
		offset = 0;
	}
}
